
package org.mule.module.hbase.processors;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.module.hbase.HbaseCloudConnector;
import org.mule.module.hbase.adapters.HbaseCloudConnectorLifecycleAdapter;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * ModifyColumnFamilyMessageProcessor invokes the {@link org.mule.module.hbase.HbaseCloudConnector#modifyColumnFamily(java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer, org.mule.module.hbase.api.CompressionType, org.mule.module.hbase.api.CompressionType, java.lang.Boolean, java.lang.Integer, java.lang.Boolean, org.mule.module.hbase.api.BloomFilterType, java.lang.Integer, java.util.Map)} method in {@link HbaseCloudConnector }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class ModifyColumnFamilyMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    private Object tableName;
    private String _tableNameType;
    private Object columnFamilyName;
    private String _columnFamilyNameType;
    private Object maxVersions;
    private Integer _maxVersionsType;
    private Object blocksize;
    private Integer _blocksizeType;
    private Object compressionType;
    private org.mule.module.hbase.api.CompressionType _compressionTypeType;
    private Object compactionCompressionType;
    private org.mule.module.hbase.api.CompressionType _compactionCompressionTypeType;
    private Object inMemory;
    private Boolean _inMemoryType;
    private Object timeToLive;
    private Integer _timeToLiveType;
    private Object blockCacheEnabled;
    private Boolean _blockCacheEnabledType;
    private Object bloomFilterType;
    private org.mule.module.hbase.api.BloomFilterType _bloomFilterTypeType;
    private Object replicationScope;
    private Integer _replicationScopeType;
    private Object values;
    private Map<String, String> _valuesType;
    private static Logger logger = LoggerFactory.getLogger(ModifyColumnFamilyMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(HbaseCloudConnectorLifecycleAdapter.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.module.hbase.adapters.HbaseCloudConnectorLifecycleAdapter"), e, this);
            }
        }
        if (moduleObject instanceof String) {
            moduleObject = muleContext.getRegistry().lookupObject(((String) moduleObject));
            if (moduleObject == null) {
                throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object by config name"), this);
            }
        }
    }

    public void start()
        throws MuleException
    {
    }

    public void stop()
        throws MuleException
    {
    }

    public void dispose() {
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets replicationScope
     * 
     * @param value Value to set
     */
    public void setReplicationScope(Object value) {
        this.replicationScope = value;
    }

    /**
     * Sets values
     * 
     * @param value Value to set
     */
    public void setValues(Object value) {
        this.values = value;
    }

    /**
     * Sets columnFamilyName
     * 
     * @param value Value to set
     */
    public void setColumnFamilyName(Object value) {
        this.columnFamilyName = value;
    }

    /**
     * Sets inMemory
     * 
     * @param value Value to set
     */
    public void setInMemory(Object value) {
        this.inMemory = value;
    }

    /**
     * Sets blocksize
     * 
     * @param value Value to set
     */
    public void setBlocksize(Object value) {
        this.blocksize = value;
    }

    /**
     * Sets tableName
     * 
     * @param value Value to set
     */
    public void setTableName(Object value) {
        this.tableName = value;
    }

    /**
     * Sets blockCacheEnabled
     * 
     * @param value Value to set
     */
    public void setBlockCacheEnabled(Object value) {
        this.blockCacheEnabled = value;
    }

    /**
     * Sets compactionCompressionType
     * 
     * @param value Value to set
     */
    public void setCompactionCompressionType(Object value) {
        this.compactionCompressionType = value;
    }

    /**
     * Sets bloomFilterType
     * 
     * @param value Value to set
     */
    public void setBloomFilterType(Object value) {
        this.bloomFilterType = value;
    }

    /**
     * Sets maxVersions
     * 
     * @param value Value to set
     */
    public void setMaxVersions(Object value) {
        this.maxVersions = value;
    }

    /**
     * Sets timeToLive
     * 
     * @param value Value to set
     */
    public void setTimeToLive(Object value) {
        this.timeToLive = value;
    }

    /**
     * Sets compressionType
     * 
     * @param value Value to set
     */
    public void setCompressionType(Object value) {
        this.compressionType = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            if (((Class) expectedType).isPrimitive()) {
                if (((Class) expectedType).getName().equals("boolean")&&(clazz == Boolean.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("byte")&&(clazz == Byte.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("short")&&(clazz == Short.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("char")&&(clazz == Character.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("int")&&(clazz == Integer.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("float")&&(clazz == Float.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("long")&&(clazz == Long.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("double")&&(clazz == Double.class)) {
                    return true;
                }
                return false;
            } else {
                return ((Class) expectedType).isAssignableFrom(clazz);
            }
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, String expectedMimeType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, expectedMimeType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, expectedMimeType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, expectedMimeType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = null;
            if (expectedMimeType!= null) {
                targetDataType = DataTypeFactory.create(((Class) expectedType), expectedMimeType);
            } else {
                targetDataType = DataTypeFactory.create(((Class) expectedType));
            }
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage _muleMessage = event.getMessage();
        HbaseCloudConnectorLifecycleAdapter _castedModuleObject = null;
        if (moduleObject instanceof String) {
            _castedModuleObject = ((HbaseCloudConnectorLifecycleAdapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (_castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("modifyColumnFamily"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            _castedModuleObject = ((HbaseCloudConnectorLifecycleAdapter) moduleObject);
        }
        try {
            String _transformedTableName = ((String) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_tableNameType").getGenericType(), null, tableName));
            String _transformedColumnFamilyName = ((String) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_columnFamilyNameType").getGenericType(), null, columnFamilyName));
            Integer _transformedMaxVersions = ((Integer) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_maxVersionsType").getGenericType(), null, maxVersions));
            Integer _transformedBlocksize = ((Integer) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_blocksizeType").getGenericType(), null, blocksize));
            org.mule.module.hbase.api.CompressionType _transformedCompressionType = ((org.mule.module.hbase.api.CompressionType) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_compressionTypeType").getGenericType(), null, compressionType));
            org.mule.module.hbase.api.CompressionType _transformedCompactionCompressionType = ((org.mule.module.hbase.api.CompressionType) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_compactionCompressionTypeType").getGenericType(), null, compactionCompressionType));
            Boolean _transformedInMemory = ((Boolean) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_inMemoryType").getGenericType(), null, inMemory));
            Integer _transformedTimeToLive = ((Integer) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_timeToLiveType").getGenericType(), null, timeToLive));
            Boolean _transformedBlockCacheEnabled = ((Boolean) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_blockCacheEnabledType").getGenericType(), null, blockCacheEnabled));
            org.mule.module.hbase.api.BloomFilterType _transformedBloomFilterType = ((org.mule.module.hbase.api.BloomFilterType) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_bloomFilterTypeType").getGenericType(), null, bloomFilterType));
            Integer _transformedReplicationScope = ((Integer) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_replicationScopeType").getGenericType(), null, replicationScope));
            Map<String, String> _transformedValues = ((Map<String, String> ) evaluateAndTransform(_muleMessage, ModifyColumnFamilyMessageProcessor.class.getDeclaredField("_valuesType").getGenericType(), null, values));
            retryCount.getAndIncrement();
            _castedModuleObject.modifyColumnFamily(_transformedTableName, _transformedColumnFamilyName, _transformedMaxVersions, _transformedBlocksize, _transformedCompressionType, _transformedCompactionCompressionType, _transformedInMemory, _transformedTimeToLive, _transformedBlockCacheEnabled, _transformedBloomFilterType, _transformedReplicationScope, _transformedValues);
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("modifyColumnFamily"), event, e);
        }
    }

}
